<?php

namespace InSegment\ApiCore\Models;

use InSegment\ApiCore\Interfaces\SADConsumer;

class SourceAttributeDownstream
{
    const ACCEPT_STATE_INITIAL = 0;
    const ACCEPT_STATE_ACCEPTED = 1;
    const ACCEPT_STATE_SAVED = 2;
    const ACCEPT_STATE_FINAL = 3;
    
    const SOURCE_ITEM_TEMPLATE = [
        'i' => 0,
        'total' => 0,
        'a' => [],
        'state' => self::ACCEPT_STATE_INITIAL,
        'params' => []
    ];
    
    const ATTRIBUTE_ITEM_TEMPLATE = [
        'i' => 0,
        'total' => 0,
        'a' => []
    ];
    
    const DOWNSTREAM_ITEM_TEMPLATE = [
        'i' => 0,
        'total' => 0,
        'a' => []
    ];
    
    /**
     * Stored rules with attributes dictionary
     *
     * @var array 
     */
    protected $classToAttributes = [];
    
    /**
     * Source. It is an array of Models which need to be processes
     * 
     * @var array 
     */
    protected $source = [
        'current' => self::SOURCE_ITEM_TEMPLATE,
        'stack' => []
    ];
    
    /**
     * Attributes. They are retrieved from the dictionary of rules and are iterated on each model
     *
     * @var array 
     */
    protected $attributes = [
        'current' => self::ATTRIBUTE_ITEM_TEMPLATE,
        'stack' => []
    ];
    
    /**
     * Downstream. In the main iteration, it is a separate branch of attributes that will be iterated after the first one
     *
     * @var array 
     */
    protected $downstream = [
        'current' => self::DOWNSTREAM_ITEM_TEMPLATE,
        'stack' => []
    ];
    
    /**
     * Constructor. Accepts attributes by class dictionary
     * 
     * @param array $rules
     */
    public function __construct(array $rules)
    {
        foreach ($rules as $class => $attributes) {
            $this->classToAttributes[$class] = self::ATTRIBUTE_ITEM_TEMPLATE;
            $this->classToAttributes[$class]['total'] = count($attributes);
            $this->classToAttributes[$class]['a'] = $attributes;
        }
    }
    
    /**
     * The main cycle. Accepts SADConsumer, array of Models, and a class for attribute rules
     * What the consumer considers 'models', 'attributes' or 'downstream' is irrelevant, also by default
     * the main cycle won't add any new sources or downstream in process, so the consumer should call
     * methods SourceAttributeDownstream::newSource and SourceAttributeDownstream::addDownStream to affect the cycle
     * 
     * @param \InSegment\ApiCore\Interfaces\SADConsumer $consumer
     * @param array $source
     * @param string $class
     * @return null
     */
    public function iterateArrayOfModels(SADConsumer $consumer, array $source, string $class)
    {
        $this->newSource($source, $class);
        // as the consumer is allowed to modify state of our cycle, and each of supplied sources must be still processed,
        // method SourceAttributeDownstream::newSource saves the previous state into the stack
        // the outer cycle handles restoration of the state saved into stack once the current source has no more entries
        do {
            // this construction allows to avoid recursion by modification of the state of the cycle
            // the while cycle there iterates through the source, which is possibly changed on almost each iteration
            while ($this->sourceHasModel()) {
                // note that on each step consumer's method can and is intended to change OUR cycle's state
                // so we cannot just hop to the next case, after each call of consumer's method we must jump to
                // the while and recheck the state
                switch ($this->source['current']['state']) {
                    // before consumer is informed of model
                    case self::ACCEPT_STATE_INITIAL:
                        $this->source['current']['state'] = self::ACCEPT_STATE_ACCEPTED;
                        $consumer->acceptModel($this->getModel(), $this->getParams());
                    break;
                    // processing of attributes step, then save of the model
                    case self::ACCEPT_STATE_ACCEPTED:
                        if ($this->hasAttribute()) {
                            $attribute = $this->getAttribute();
                            $this->advanceAttribute();
                            $consumer->acceptAttribute($attribute);
                        } else if ($this->hasDownstream()) {
                            $downstream = $this->getDownstream();
                            $this->advanceDownstream();
                            $consumer->acceptDownstream($downstream);
                        } else {
                            $this->source['current']['state'] = self::ACCEPT_STATE_SAVED;
                            $consumer->saveModel($this->getModel(), $this->getParams());
                        }
                    break;
                    // the model is saved, processing downstream relations
                    case self::ACCEPT_STATE_SAVED:
                        $this->advanceSource();
                    break;
                }
            }
        } while ($this->unstackSource());
    }
    
    /**
     * Add new array of Models so it will become the current source and stack everything down
     * 
     * @param array $source
     * @param string $class
     * @param array $params
     * @return null
     */
    public function newSource(array $source, string $class, array $params = [])
    {
        $this->attributes['stack'][] = $this->attributes['current'];
        $this->attributes['current'] = $this->classToAttributes[$class];
        
        $this->downstream['stack'][] = $this->downstream['current'];
        $this->downstream['current'] = self::DOWNSTREAM_ITEM_TEMPLATE;
        
        $this->source['stack'][] = $this->source['current'];
        $this->source['current'] = self::SOURCE_ITEM_TEMPLATE;
        $this->source['current']['params'] = $params;
        $this->source['current']['total'] = count($source);
        $this->source['current']['a'] = &$source;
    }
    
    /**
     * Add something to the end of the downstream
     * 
     * @param mixed $value
     */
    public function addDownStream($value)
    {
        $this->downstream['current']['a'][] = $value;
        ++$this->downstream['current']['total'];
    }
    
    /**
     * Iterate source to the next value
     * Reset attribute iteration on the level
     * 
     * @return null
     */
    protected function advanceSource()
    {
        ++$this->source['current']['i'];
        $this->source['current']['state'] = self::ACCEPT_STATE_INITIAL;
        $this->attributes['current']['i'] = 0;
    }
    
    /**
     * Check validity of source iteration
     * 
     * @return bool
     */
    protected function sourceHasModel()
    {
        return $this->source['current']['i'] < $this->source['current']['total'];
    }
    
    /**
     * Ladder up the stack if possible, return true if successful, false if the stack was empty
     * 
     * @return boolean
     */
    protected function unstackSource()
    {
        if (empty($this->source['stack'])) {
            return false;
        }
        
        $this->source['current'] = array_pop($this->source['stack']);
        $this->downstream['current'] = array_pop($this->downstream['stack']);
        $this->attributes['current'] = array_pop($this->attributes['stack']);
        return true;
    }
    
    /**
     * Get model from source
     * Validity of the call must be checked beforehand
     * 
     * @return mixed
     */
    protected function getModel()
    {
        return $this->source['current']['a'][$this->source['current']['i']];
    }
    
    /**
     * Get params for consumer calls
     * 
     * @return array
     */
    protected function getParams()
    {
        return $this->source['current']['params'];
    }
    
    /**
     * Iterate attribute to the next value
     * 
     * @return null
     */
    protected function advanceAttribute()
    {
        ++$this->attributes['current']['i'];
    }
    
    /**
     * Check validity of attribute iteration
     * 
     * @return bool
     */
    protected function hasAttribute()
    {
        return $this->attributes['current']['i'] < $this->attributes['current']['total'];
    }
    
    /**
     * Get attribute
     * Validity of the call must be checked beforehand
     * 
     * @return mixed
     */
    protected function getAttribute()
    {
        return $this->attributes['current']['a'][$this->attributes['current']['i']];
    }
    
    /**
     * Iterate downstream to the next value
     * 
     * @return null
     */
    protected function advanceDownstream()
    {
        ++$this->downstream['current']['i'];
    }
    
    /**
     * Check validity of downstream iteration
     * 
     * @return bool
     */
    protected function hasDownstream()
    {
        return $this->downstream['current']['i'] < $this->downstream['current']['total'];
    }
    
    /**
     * Get downstream
     * Validity of the call must be checked beforehand
     * 
     * @return mixed
     */
    protected function getDownstream()
    {
        return $this->downstream['current']['a'][$this->downstream['current']['i']];
    }
}
